#!/bin/bash
#
# Build packages for multiple arches and dists from a spec file.
#
# Requires Fedora packages:
#   mock (see patch notes below)
#   repoview
#
# The user running this script must be a member of group 'mock'.
#
# Mock patch
# --
#
# As written this script requires a patch to mock. See:
#
#    https://github.com/ringerc/mock/tree/target-paths
#    https://bugzilla.redhat.com/show_bug.cgi?id=1126117
#
# Mock is trivial to rebuild, just clone it from git and
# 'make rpm'
#  
# If you want to use it with an unpatched mock, copy everything in mock-targets/
# to /etc/mock, and invoke with:
#
#     targets="centos-5-i386 centos-5-x86_64 centos-6-i386 centos-6-x86_64 centos-7-x86_64 fedora-19-i386 fedora-19-x86_64 fedora-20-i386 fedora-20-x86_64" ./mock-build-all.sh
#
# Building just some packages
# ---
#
# To build just a subset of packages:
#
#     targets="mock-targets/fedora-20-x86_64.cfg" ./mock-build-all.sh
#
# Currently if you want to build the i386 ones you must build the x64 variant
# too.
#
# Why isn't this a Makefile?
# ----
# 
# Mainly because it's a pain encoding mock's rules about the output RPM
# name(s), which depend on the spec file contents, as dependency rules
# and targets.
#
# Also, because for general use this'll probably run as a bunch of
# Jenkins jobs anyway.
#
# Predicting output files
# ----
#
# The files needed by a spec may be listed with:
#
#  spectool -lf unified-rpm/postgresql94.spec
#
# To get the product SRPM names:
#  
#  rpm -q --qf '%{name}-%{version}-%{release}.srpm\n' --specfile unified-rpm/postgresql94.spec | head -1
#
# (but note that you'd have to set %{dist} correctly on the rpm cmdline or run under mock to
# get a correct result. Mock has the 'dist' tag in its config files.)
#
# To get the product RPM names:
#
#  rpm -q --qf "$(rpm --eval '%{_build_name_fmt}')\n" --specfile unified-rpm/postgresql94.spec
# 
# but again you have to define %{dist} . Oddly you can't seem to use `-p` and a
# specfile instead; it only lists the base package name.
#

set -e -u
shopt -s nullglob

mkdir -p mock
mkdir -p repo

SPECFILE="unified-rpm/postgresql94.spec"
SOURCES="unified-rpm/"

MOCK_EXTRA_OPTS=${MOCK_EXTRA_OPTS:-}

# The version suffix for the PGDG repo name eg pgdg94, and
# package name eg postgresql94.
PGDGVERSION=${PGDGVERSION:-94}

DIE_ON_FIRST_ERROR=1
    
export RPM_BUILD_NCPUS=${RPM_BUILD_NCPUS:-4}
PARALLEL=${PARALLEL:-1}

default_targets="
    mock-targets/centos-5-i386.cfg
    mock-targets/centos-5-x86_64.cfg
    mock-targets/centos-6-i386.cfg
    mock-targets/centos-6-x86_64.cfg
    mock-targets/centos-7-x86_64.cfg
    mock-targets/fedora-19-i386.cfg
    mock-targets/fedora-19-x86_64.cfg
    mock-targets/fedora-20-i386.cfg
    mock-targets/fedora-20-x86_64.cfg
"


targets=${targets:-$default_targets}

get-dist() {
    # Get the distro name sans path, suffix, and arch
    basename ${1%.cfg}
}

get-repoview-title() {
    dist=$(echo $1 |cut -d '-' -f 1)
    rel=$(echo $1 |cut -d '-' -f 2)
    arch=$(echo $1 | cut -d '-' -f 3)
    if [ "$dist" = "centos" ]; then
        distname="CentOS"
    elif [ "$dist" = "fedora" ]; then
        distname="Fedora"
    else
        echo "unknown dist $dist"
        exit 1
    fi
    if [ -z "$arch" ]; then
        arch="Sources"
    fi
    echo "$distname $rel $arch"
}

# SRPMs
#
# if you want to avoid rebuilding for each target os, you can instead
# pass 
#
#    --define "_source_filedigest_algorithm md5" \ 
#    --define "_binary_filedigest_algorithm md5" 
#
# to mock on the cmdline, but I prefer to make SRPMs that're built on the
# target directly.
#

build-srpm() {
    distrelease=$(get-dist $target | cut -d '-' -f '1-2')
    resultdir="mock/$distrelease"
    echo "Building srpm for $resultdir"
    rm -rf "$resultdir"
    if ! mock ${MOCK_EXTRA_OPTS} -q -r $target --buildsrpm --spec "$SPECFILE" --sources "$SOURCES" --resultdir "$resultdir"
    then
        echo "ERROR: srpm build failure for $target; see mock logs in $resultdir for details"
        if [ $DIE_ON_FIRST_ERROR ]; then exit 1; fi
    else
        echo "Adding $distrelease srpm to repo"
        mkdir -p repo/$distrelease
        find "$resultdir" -name \*.src.rpm -exec cp {} repo/$distrelease/ \;
        createrepo -q repo/$distrelease/
        repoview -q -t "$(get-repoview-title $distrelease)" repo/$distrelease/
    fi
    echo "Done building SRPM for $distrelease"
}

build-rpms() {
    distreleasearch=$(get-dist $target)
    distrelease=$(get-dist $target | cut -d '-' -f '1-2')
    srpm=$(find "mock/$distrelease" -name \*.src.rpm)
    if [ -z "$srpm" ]; then
        echo "ERROR: srpm not found for target $target - earlier build failure?"
        continue
    fi
    echo "Rebuilding $srpm for $target"
    echo "Building $target"
    resultdir="mock/$distreleasearch"
    if ! mock ${MOCK_EXTRA_OPTS} -q -r $target --rebuild "$srpm" --resultdir "$resultdir"
    then
        echo "ERROR: rpm build failure for $target, see mock logs in $resultdir for details"
        if [ $DIE_ON_FIRST_ERROR ]; then exit 1; fi
    else
        echo "Adding ${distreleasearch} RPMs to repo"
        mkdir -p repo/${distreleasearch}
        basearch=$(echo $distreleasearch | cut -d '-' -f 3)
        find "$resultdir" -name \*.${basearch}.rpm -exec cp {} repo/$distreleasearch/ \;
        createrepo -q  repo/$distreleasearch/
        repoview -q -t "$(get-repoview-title $distreleasearch)" repo/$distreleasearch/
    fi
    echo "Done building RPMs for $distreleasearch"
}

check-rpm() {
    distreleasearch=$(get-dist $target)
    rpms=$( find "mock/$distreleasearch" -name \*.rpm )
    echo "Testing installation in $target"
    # Create a new resultdir so we preserve the old one's packages etc
    mock -q -r $target --resultdir "mock/${target#*/}-rpmtest" --init
    # then install the packages
    mock -q -r $target --resultdir "mock/${target#*/}-rpmtest" --install $rpms
    # not much more testing is possible, because mock is a chroot, so there
    # are port conflicts to deal with, plus systemd throws a giant wobbly.
    echo "Install checks for $distreleasearch done"
}

spectool --gf "$SPECFILE" --directory "unified-rpm"

for target in $targets
do
    if [[ "$target" != *x86_64* ]] ; then
        # No need to build SRPMs for x86_64 and other arches, and all our
        # targets support x86_64
        continue
    fi
    if [ "$PARALLEL" -eq 1 ]; then
        build-srpm "$target" &
    else
        build-srpm "$target"
    fi
done
if [ "$PARALLEL" -eq 1 ]; then
    echo "Waiting for SRPM builds to finish..."
    # plain 'wait' always returns 0, so:
    for j in $(jobs -p) ; do
        wait $j
    done
    echo "SRPM builds finished"
fi

# RPMs
for target in $targets
do
    if [ "$PARALLEL" -eq 1 ]; then
        build-rpms "$target" &
    else
        build-rpms "$target"
    fi
done
if [ "$PARALLEL" -eq 1 ]; then
    echo "Waiting for RPM builds to finish..."
    # plain 'wait' always returns 0, so:
    for j in $(jobs -p) ; do
        wait $j
    done
    echo "done"
fi

# Test installation of each RPM
#
# This'll try to install all RPMs in a mock dir, which might cause issues if there were
# multiple variants of the same package produced.
#
# TODO: It'd be better to generate the RPM file names (as above) and use them.
#
for target in $targets
do
    check-rpm "$target"
done

if git status >&/dev/null
then
cat > repo/README <<__END__
This repository was generated by mock-build-all.sh from the packaging
repository, with:

branch: $(git rev-parse --abbrev-ref HEAD || echo 'detached HEAD')
commit hash: $(git rev-parse --short HEAD || echo 'unknown')
__END__
fi
